import java.io.*;
import java.util.*;

/**
 * KruskalWallis - a Java utility to perform a Kruskal-Wallis test on a table of data
 * <p>
 * 
 * The Kruskal-Wallis test is a non-parametric statistical test used to assess the difference
 * between three or more independent samples, which is to say, between three or more conditions
 * assigned between-subjects. The relationship with other non-parametric tests is illustrated below:
 * <p>
 * 
 * <blockquote><img src="KruskalWallis-0.jpg"></blockquote>
 * <p>
 * 
 * The test is also called a one-way analysis of variance <i>by ranks</i>. If the test is used with
 * interval or ratio-scale data, the data are first converted to ranks. Using the Kruskal-Wallis
 * test for interval or ratio data might be preferred over a parametric ANOVA if there is a concern
 * over the underlying assumptions in the distribution of the data. The researcher must bear in
 * mind, however, that the data transformation bears a cost. Information is sacrificed. For example,
 * interval or ratio scores of 10, 49, and 50 are converted to ranks 1, 2, and 3.
 * <p>
 * 
 * The null hypothesis is that the populations from which the samples originate have the same
 * median. A statistically significant outcome means that at least one of the samples (i.e.,
 * conditions) is significantly different from at least one other sample. To determine which pairs
 * of samples differ significantly from one another, a post hoc pairwise comparisons test is
 * performed (see below).
 * <p>
 * 
 * Invocation (usage message if invoked without arguments):
 * <p>
 * 
 * <blockquote><a href="KruskalWallis-2.jpg"><img src="KruskalWallis-2.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * <b>Example.</b> Consider an experiment to test whether age will be a factor in the acceptance of
 * a new GPS system. Eight participants are recruited from each of three age groups (20-29, 30-39,
 * 40-49). The participants are given a demo of a new GPS system, then asked if they liked it enough
 * to purchase it for personal use. They respond on a 10-point linear scale (1= definitely no, 10 =
 * definitely yes). The responses are saved in <a
 * href="kruskalwallis-ex1.txt">kruskalwallis-ex1.txt</a> as a matrix with 8 rows and 3 columns. The
 * data are then processed by the <code>KruskalWallis</code> utility. For example,
 * 
 * <blockquote><a href="KruskalWallis-3.jpg"><img src="KruskalWallis-3.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * The test statistic H follows the chi-square distribution with <i>k</i> - 1 degrees for freedom
 * (<i>k</i> = number of conditions). Since <i>p</i> is below the customary threshold for
 * significance (.05), we reject the null hypothesis. The conclusion is that age is a factor in
 * acceptance of the new design for a GPS system.
 * <p>
 * 
 * The test also provides <i>H'</i> which is corrected for ties. The difference is usually minor.
 * <p>
 * 
 *  StatView provides the following output using the data above:
 * <p>
 * 
 * <blockquote><img src="KruskalWallis-1.jpg"></blockquote>
 * <p>
 * 
 * <b>Post hoc comparisons.</b> If the test reveals statistical significance (as above), a post hoc
 * pairwise comparisons test is needed to determine which pairs of conditions (samples) differ
 * significantly from one another. This is available using the <code>-ph</code> option:
 * <p>
 * 
 * <blockquote><a href="KruskalWallis-4.jpg"><img src="KruskalWallis-4.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * The test reveals that only the samples in columns 1 and 3 are significantly different. The
 * verbose output (see below) provides a rough verification of this, since the means for columns 1
 * and 3 were 18.5 and 7.9, respectively.
 * <p>
 * 
 * The multiple comparisons test uses the procedure given by Siegel and Castellan in
 * "Nonparametric Statistics for the Behavioral Sciences" (McGraw-Hill, 1988, pp. 213-214; described
 * in a posting on www.talkstats.com).
 * <p>
 * 
 * <b>Missing data.</b> An additional feature of the <code>KruskalWallis</code> utility is
 * accommodation for missing data. If the sample sizes (number of participants) differ among the
 * conditions, substitute a period (".") for missing data. An example is found in <a
 * href="kruskalwallis-ex2.txt">kruskalwallis-ex2.txt</a>:
 * <p>
 * 
 * <blockquote><a href="KruskalWallis-5.jpg"><img src="KruskalWallis-5.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * The example above contains scores from a study on wine tasting (see
 * http://vassarstats.net/textbook/ch14a.html).
 * <p>
 * 
 * @author Scott MacKenzie, 2012-2014
 */
public class KruskalWallis
{
	public static void main(String[] args) throws IOException
	{
		boolean vOption = false;
		boolean phOption = false;

		if (args.length < 1)
		{
			System.out.println("------------------------------------------------------------------------\n"
					+ "Usage: java KruskalWallis file [-ph] [-v]\n" + "\n"
					+ "  where \'file\' contains a data table (comma or space delimited)\n"
					+ "        [-ph]  = include post hoc pairwise comparisons test\n"
					+ "        [-v]   = verbose (debugging and other information)\n"
					+ "------------------------------------------------------------------------\n");
			System.exit(0);
		}

		String file = args[0];

		for (int i = 1; i < args.length; ++i)
		{
			if (args[i].equals("-v"))
				vOption = true;
			if (args[i].equals("-ph"))
				phOption = true;
		}

		BufferedReader br = new BufferedReader(new FileReader(file));
		String s = null;
		Vector<String> v = new Vector<String>();
		double[][] data;
		StringTokenizer st;
		int numberOfColumns = -1;

		// read the data
		while ((s = br.readLine()) != null)
			v.add(s);

		// determine number of rows and columns
		int nRows = v.size();
		String tmp = v.elementAt(0);
		int nCols = (new StringTokenizer(tmp)).countTokens();

		if (nCols < 3)
		{
			System.out.println("Data format error.  At least three columns required.");
			System.exit(0);
		}

		// create data table
		data = new double[nRows][nCols];

		// fill data table
		for (int i = 0; i < data.length; ++i)
		{
			st = new StringTokenizer(v.elementAt(i));
			if (numberOfColumns == -1)
				numberOfColumns = st.countTokens();
			else if (numberOfColumns != nCols)
			{
				System.out.println("Error: Data format error.  Columns = " + st.countTokens());
				System.exit(0);
			}

			int j = 0;
			while (st.hasMoreTokens())
			{
				String ss = st.nextToken();
				if (ss.equals("."))
					data[i][j++] = Double.NaN;
				else
					data[i][j++] = Double.parseDouble(ss);
			}
		}

		if (vOption)
		{
			System.out.println("Raw data...");
			for (int i = 0; i < data.length; ++i)
			{
				for (int j = 0; j < data[0].length; ++j)
					System.out.print(data[i][j] + ", ");
				System.out.println();
			} // data table loaded!
			System.out.println();
		}

		double[] tmp2 = new double[nRows * nCols];
		int countNaN = 0;
		for (int i = 0; i < tmp2.length; ++i)
		{
			tmp2[i] = data[i % nRows][i / nRows];
			if (Double.isNaN(tmp2[i]))
				++countNaN;
		}

		double[] tmp3 = new double[tmp2.length - countNaN];
		for (int i = 0, j = 0; i < tmp2.length; ++i)
			if (!Double.isNaN(tmp2[i]))
				tmp3[j++] = tmp2[i];

		double[] ranks = new double[tmp3.length];
		ranks = getRanks(tmp3);

		double[][] rankData = new double[nRows][nCols];
		for (int i = 0, j = 0; i < nRows * nCols; ++i)
			if (!Double.isNaN(tmp2[i]))
				rankData[i % nRows][i / nRows] = ranks[j++];
			else
				rankData[i % nRows][i / nRows] = Double.NaN;

		double[] count = new double[nCols];
		double[] total = new double[nCols];
		for (int i = 0; i < nCols; ++i)
			for (int j = 0; j < nRows; ++j)
				if (!Double.isNaN(rankData[j][i]))
				{
					++count[i];
					total[i] += rankData[j][i];
				}

		double[] mean = new double[nCols];
		for (int i = 0; i < mean.length; ++i)
			mean[i] = total[i] / count[i];

		double[] t2Count = new double[nCols];
		for (int i = 0; i < t2Count.length; ++i)
			t2Count[i] = total[i] * total[i] / count[i];

		int n = ranks.length;

		double A = 12.0 / (n * (n + 1));
		double B = 0.0;
		for (int i = 0; i < t2Count.length; ++i)
			B += t2Count[i];
		double C = 3 * (n + 1);

		if (vOption)
		{
			printArray(mean, "column means = ");
			System.out.println();

			System.out.println("Rank data...");
			for (int i = 0; i < rankData.length; ++i)
			{
				for (int j = 0; j < rankData[0].length; ++j)
					System.out.print(rankData[i][j] + ", ");
				System.out.println();
			}
			System.out.println();

			printArray(count, "counts = ");
			printArray(total, "totals = ");
			printArray(t2Count, "t2Counts = ");
			System.out.println("n = " + n);
			System.out.println("A = " + A + ", B = " + B + ",  C= " + C);
			System.out.println();
		}

		double h1 = A * B - C;
		double p1 = Statistics.chiSquaredProbability(h1, nCols - 1);
		System.out.printf("H = %.3f, p = %.4f\n", h1, p1);

		double factor = correctionForTies(ranks);
		double cft = 1.0 - factor / (n * n * n - n);

		double h2 = h1 / cft;
		double p2 = Statistics.chiSquaredProbability(h2, nCols - 1);
		System.out.printf("H' = %.3f, p' = %.4f\n", h2, p2);

		// --------------------------------
		// ----- POST HOC COMPARISONS -----
		// --------------------------------
		// This calculation is based on a posting to www.talkstats.com

		if (phOption)
		{
			System.out.println("\n----------------------------------------------------");
			System.out.println("----- Multiple Comparisons Test (alpha = .05) ------");
			System.out.println("----------------------------------------------------");
			for (int i = 0; i < nCols - 1; ++i)
			{
				for (int j = i + 1; j < nCols; ++j)
				{
					double r1, r2, n1, n2, nn;
					r1 = r2 = n1 = n2 = nn = 0.0;
					for (int k = 0; k < nRows; ++k)
					{
						if (!Double.isNaN(rankData[k][i]))
						{
							r1 += rankData[k][i];
							++n1;
							++nn;
						}
						if (!Double.isNaN(rankData[k][j]))
						{
							r2 += rankData[k][j];
							++n2;
							++nn;
						}
					}
					r1 = r1 / n1;
					r2 = r2 / n2;
					double ls = Math.abs(r1 - r2);

					double k = 9.0;
					double zcrit = Statistics.normalInverse(0.05 / (k * (k - 1)));
					double sqrt = Math.sqrt(((nn * (nn + 1)) / 12) * (1 / n1 + 1 / n2));
					double rs = Math.abs(zcrit * sqrt);

					String decision = (ls >= rs) ? "* (significant)" : "-";
					System.out.printf("Pair %d:%d --> %7.4f >= %7.4f  ?   %s\n", (i + 1), (j + 1), ls, rs, decision);

				}
			}
			System.out.println("----------------------------------------------------");
		}
	}

	// given an array of data, returns a array of ranks (handles ties)
	private static double[] getRanks(double[] d)
	{
		// put raw ranks in 'r' array
		double[] r = new double[d.length];

		// create a copy of 'd' array
		double[] a = new double[d.length];
		for (int i = 0; i < a.length; ++i)
			a[i] = d[i];
		Arrays.sort(a); // sort, but leave values intact

		for (int i = 0; i < a.length; ++i)
		{
			double rawScore = a[i];
			int rawRank = i + 1; // raw rank
			int count = 1;
			int j = i + 1;
			while (j < a.length) // look for ties (potentially to end of array)
			{
				if (rawScore == a[j]) // found a tie
				{
					rawRank += j + 1;
					++count;
					++j;
				} else
					// no tie, skip to next position in array
					break;
			}
			double rank = (double)rawRank / count;
			r[i] = rank;
			while (--count > 0) // continue to fill slots with ties
			{
				++i;
				r[i] = rank;
			}
		}

		// use a (raw scores, sorted), r (ranks, sorted), and d (raw scores, unsorted) to build rr
		double[] rr = new double[d.length];

		for (int i = 0; i < d.length; ++i)
		{
			for (int j = 0; j < d.length; ++j)
			{
				if (d[i] == a[j])
				{
					rr[i] = r[j];
					break;
				}
			}
		}
		return rr;
	}

	private static double correctionForTies(double[] rArg)
	{
		double factor = 0.0;

		double[] r = new double[rArg.length];
		for (int i = 0; i < r.length; ++i)
			r[i] = rArg[i];
		Arrays.sort(r);

		for (int i = 0; i < r.length; ++i)
		{
			double rank = r[i];
			int j = i + 1;
			int countRanks = 1;
			while (j < r.length)
			{
				if (rank == r[j])
				{
					++countRanks;
					++i;
					++j;
				} else
					break;
			}
			if (countRanks > 1)
				factor += countRanks * countRanks * countRanks - countRanks;
		}
		return factor;
	}

	private static void printArray(double[] d, String s)
	{
		System.out.print(s);
		for (int i = 0; i < d.length; ++i)
			System.out.print(d[i] + ", ");
		System.out.println();
	}
}
